;----------------------------------------------------------------------------
; KERNEL (RELOCATION PROGRAM)
;----------------------------------------------------------------------------
exec:
        movem.l d0-d7/a0-a6,-(a7)
	; TIE get's the handle of the program in an internal variable
        
        lea     Prgm(PC),a0
        move.w  (a0),d1

	lsl.w	#2,d1			; Get file address
        move.l  doorsos::heap,a4
	move.l	0(a4,d1.w),a4

        lea     2(a4),a6                ; Skip size
        bsr     RelocFromPtr            ; do relocation
        tst.l   d0                      ; successful
        bne     execError
	lea	Prgm(PC),a0
	move.w  (a0),-(a7)
        lea.l   ProgCount(pc),a0
        tst.w   (a0)
        bne     execNoSaveScreen

;-----------------------------------------------------------------------------
; Register Anti Crash protection
	move.w	$E(a6),d0			; Get user exit address
	lea	UserExit-4(PC),a3
	lea	CrashStack-4(PC),a4
	movea.l	#0,a1
	tst.w	d0
	beq	RegisterExit
	lea	0(a6,d0.w),a1
RegisterExit:
	; Find space
	lea	4(a3),a3
	lea	4(a4),a4
	tst.l	(a4)
	bne	RegisterExit

	move.l	a1,(a3)				; User exit
	move.l	a7,(a4)

	move.w	#$700,d0
	trap	#1
	bclr.b	#2,($600001)
	lea	CrashHandler(PC),a1
	lea	CrashTable(PC),a2
	clr.l	d3
InstallHandler:
	move.b	(a2)+,d3
	tst.b	d3
	beq	CrashProtectionInstalled
	move.l  d3,a3
	move.l	a1,(a3)
	bra	InstallHandler
CrashProtectionInstalled:
	bset.b	#2,($600001)
	trap	#1

        move.l  #3840,-(a7)             ; save screen
        jsr     doorsos::HeapAlloc
        add.l   #4,a7
        tst.w   d0                      ; got memory
        bne     execCopyScreen

        moveq.l #1,d0                   ; out of memory
        bra     execError

execCopyScreen:				; Save screen
        lea.l   ScreenHandle(pc),a0
        move.w  d0,(a0)
	
	lsl.w	#2,d0
	move.l  doorsos::heap,a0
	move.l	0(a0,d0.w),a0
;        _DEREF d0,a0
        move.l  #LCD_MEM,a1		; Screen address
        move.w  #959,d0
execSaveLoop:
        move.l (a1)+,(a0)+
        dbra d0,execSaveLoop

execNoSaveScreen:
        lea.l   ProgCount(pc),a0        ; actualize internal program counter
        add.w   #1,(a0)
        clr.l   d0

;-----------------------------------------------------------------------------
        move.w  $c(a6),d1		; Get program entry
        add.l   d1,a6

        jsr     (a6)			; call the relocated program

	bra	UnRelocProgram
ProgramCrashed:
	; Restore original stack address
	lea	ProgCount-4(PC),a0
	lea	CrashStack-4(PC),a1
	lea	UserExit(PC),a2
GetStackAddress:
	lea	4(a2),a2
	lea	4(a1),a1
	tst.l	(a1)
	bne	GetStackAddress
	movea.l -4(a1),a7		; Get new stack address
	movea.l -4(a2),a5
	cmpa.l	#0,a5
	beq	UnRelocProgram
	movem.l	d0-d7/a0-a6,-(a7)
	jsr	(a5)			; Call user exit point
	movem.l	(a7)+,d0-d7/a0-a6

UnRelocProgram:

	; Kill last crash entries
	lea	CrashStack(PC),a1
	lea	UserExit(PC),a2
GetStackAddr:
	lea	4(a2),a2
	tst.l	(a1)+
	bne	GetStackAddr
	clr.l	-8(a1)			; Kill that stuff
	clr.l	-8(a2)

	move.w	(a7)+,d1		; Get program entry
        lsl.w   #2,d1
        move.l  doorsos::heap,a4
	move.l	0(a4,d1.w),a4

        lea     2(a4),a6                ; Skip size

        bsr     UnrelocFromPtr		; unrelocate the program

        lea.l   ProgCount(pc),a0	; Restore screen?
        sub.w   #1,(a0)
        bne     execEnd
        lea.l   ScreenHandle(pc),a0     ; restore original screen
        move.w  (a0),d0
        move.w  d0,d1

	lsl.w	#2,d0
	move.l  doorsos::heap,a0
	move.l	0(a0,d0.w),a0

        move.l  #LCD_MEM,a1		; screen address
        move.w  #959,d0
execRestoreLoop:
        move.l  (a0)+,(a1)+
        dbra    d0,execRestoreLoop
        move.w  d1,-(a7)
        jsr     doorsos::HeapFree
        add.l   #2,a7

	; Unregister Crashprotection
	move.w	#$700,d0
	trap	#1

	bclr.b	#2,($600001)
	lea	CrashTable(PC),a2
	move.l	#ROMBASE,a1
	clr.l	d1
RemoveCrashProtection:
	move.b	(a2)+,d1
	tst.b	d1
	beq	UninstallDone
	move.l	d1,a3
	move.l	0(a1,d1.w),(a3)
	bra	RemoveCrashProtection
UninstallDone:
	bset.b	#2,($600001)
	trap	#1

execEnd:
	lea	CrashID(PC),a0
	tst.w	(a0)
	beq	CrashBack
	moveq	#6,d0
	bra	execError
CrashBack:
	clr.b	tios::OSOnBreak		; No break!
        movem.l (a7)+,d0-d7/a0-a6
        rts

execError:
	; Test some other stuff
	lea	ProgCount(PC),a4
	tst.w	(a4)
	bne	execEnd			; Don't show any message in the st-line
        cmp	#1,d0
        beq     execNoMem
        cmp	#2,d0
        beq     execNoLib
        cmp	#3,d0
        beq     execLibVer
	cmp	#4,d0
	beq	exectie
        cmp     #5,d0
        beq     execlib
	cmp	#6,d0
	beq	crashed
        pea.l   BadFileFormatMsg(pc)
        jsr     doorsos::ST_showHelp
        add.l   #4,a7
        bra	execEnd
execNoMem:
        pea.l   NoMemMsg(pc)
        jsr     doorsos::ST_showHelp
        add.l   #4,a7
        bra     execEnd
execNoLib:
        lea.l   MissingLibName(pc),a1
        move.l  (a0)+,(a1)+
        move.l  (a0)+,(a1)+
        pea.l   MissingLibMsg(pc)
        jsr     doorsos::ST_showHelp              ; st_showhelp
        add.l   #4,a7
        bra     execEnd
execLibVer:
        lea.l   WrongLibVersionName(pc),a1
        move.l  (a0)+,(a1)+
        move.l  (a0)+,(a1)+
        pea.l   WrongLibVersionMsg(pc)
        jsr     doorsos::ST_showHelp          ; st_showhelp
        add.l   #4,a7
        bra     execEnd
exectie:
	pea	tietooold(pc)
        jsr     doorsos::ST_showHelp         ; st_showhelp
	add.l	#4,a7
	bra	execEnd
execlib:
        pea     libnotexecutable(pc)
        jsr     doorsos::ST_showHelp
        add.l   #4,a7
        bra     execEnd
crashed:
	clr.w	(a0)				; a0 -> pointer to CrashID
	pea	CrashHappened(pc)
	jsr	doorsos::ST_showHelp
	add.l	#4,a7
	bra	CrashBack


Reloc:
        movem.l d1-d7/a1-a6,-(a7)
	
	lsl.w	#2,d0
        move.l  doorsos::heap,a6
	move.l	0(a6,d0.w),a6
;        _DEREF d0,a6

        bra RelocFromPtrStart

RelocFromPtr:
        movem.l d1-d7/a1-a6,-(a7)
RelocFromPtrStart:
        move.l  a6,a4
        tst.w   $8(a6)
        bne     RelocDone
        clr.w   $12(a6)
        move.l  4(a4),d0
        and.l   #$ffffff00,d0
        cmp.l   #$36386b00,d0
        bne     RelocBadFileFormat
        clr.l   d0
        move.w  $14(a4),d0
        tst.w   d0
        beq     RelocNoBSSAlloc
        add.l   d0,a4
        tst.l   (a4)
        beq     RelocNoBSSAlloc
        move.l  (a4),d0
        movem.l a4-a6,-(a7)
        move.l  d0,-(a7)
        jsr     doorsos::HeapAlloc              ; heapAlloc
        add.l   #4,a7
        movem.l (a7)+,a4-a6
        tst.w   d0
        beq     RelocNoMem
        move.w  d0,$12(a6)
RelocNoBSSAlloc:
        move.l  a6,a4
        add.l   #$1a,a4
        move.w  (a4)+,d7
RelocLibCheckLoop:
        tst.w   d7
        beq     RelocLibCheckEnd
        move.l  a4,a0
        bsr     GetFileAddress
        tst.w   d0
        beq     RelocLibNotFound
        move.l  6(a0),d0
        and.l   #$ffffff00,d0
        cmp.l   #$36386b00,d0
        bne     RelocLibNotFound
        add.l   #10,a4
        dbra    d7,RelocLibCheckLoop
RelocLibCheckEnd:
        move.l  a6,a4
        add.l   #$1a,a4
        move.w  (a4)+,d7
        move.w  d7,d6
        mulu.w  #10,d6
        move.l  a4,a3
        add.l   d6,a3
RelocLibLoop:
        tst.w   d7
        beq     RelocLibEnd
        move.l  a4,a0
        bsr     GetFileAddress
        tst.w   d0
        beq     RelocLibNotFound
        move.w  d0,d1
        movem.l d1-d7/a0-a6,-(a7)
        move.l  a0,a6
        add.l   #2,a6
        bsr     RelocFromPtr
        movem.l (a7)+,d1-d7/a0-a6
        tst.w   d0
        bne     RelocLibNotFound

	lsl.w	#2,d1
        move.l  doorsos::heap,a2
	move.l	0(a2,d1.w),a2
;        _DEREF d1,a2

        add.l   #2,a2
        move.l  a2,a5
        clr.l   d0
        move.w  $16(a2),d0
        add.l   d0,a2
        move.w  #1,d3
        bsr     DoReloc
        tst.w   d0
        bne     RelocWrongLibVersion
        add.l   #10,a4
        dbra    d7,RelocLibLoop
RelocLibEnd:
        tst.w   (a3)+
        beq     RelocNoROM

        move.l  $c8,a2
        sub.w	#2,a2
        move.w	#2,d3
        bsr 	DoReloc
	tst.w	d0
	bne	RelocWrongtieVersion

RelocNoROM:
        tst.w   (a3)+
        beq     RelocNoRAM
        clr.l   d0
        move.w  $18(a6),d0
        move.l  a6,a4
        add.l   d0,a4
        lea.l   _CALCULATOR(pc),a0
        move.b  #CALC_TI,(a0)
        lea.l   RAMTable92(pc),a2
        lea.l   2(a4),a4
RelocDoRAM:            
        move.w  #2,d3
;	lea	4(a6),a6
        bsr     DoReloc
;	lea	-4(a6),a6
RelocNoRAM:
        clr.l   d0
        move.w  (a3)+,d0
        tst.w   d0
        beq     RelocRelEnd
        move.l  a6,a0
        add.l   d0,a0
        move.l  a6,d0
        add.l   d0,(a0)
        bra     RelocNoRAM
RelocRelEnd:
        tst.w   $14(a6)
        beq     RelocDone
        tst.w   $12(a6)
        beq     RelocDone
        add.l   #4,a3
        move.w  $12(a6),d0

	lsl.w	#2,d0
	move.l  doorsos::heap,a1
	move.l	0(a1,d0.w),a1

RelocBSSLoop:
        clr.l   d0
        move.w  (a3)+,d0
        tst.w   d0
        beq     RelocDone
        move.l  a6,a0
        add.l   d0,a0
        move.l  a1,d0
        add.l   d0,(a0)
        bra     RelocBSSLoop
RelocDone:
        add.w   #1,$8(a6)
RelocOK:
        clr.l   d0
RelocEnd:
        movem.l (a7)+,d1-d7/a1-a6
        rts

RelocBadFileFormat:
        moveq.l #4,d0
        bra     RelocEnd
RelocLibNotFound:
        bsr     DeallocBSS
        move.l  a4,a0
        moveq.l #2,d0
        bra     RelocEnd
RelocNoMem:
        moveq.l #1,d0
        bra     RelocEnd
RelocWrongLibVersion:
        bsr     DeallocBSS
        move.l  a4,a0
        moveq.l #3,d0
        bra     RelocEnd
RelocWrongtieVersion:
	bsr	DeallocBSS
	move.l	a4,a0
	moveq	#4,d0
	bra	RelocEnd

Unreloc:
        movem.l d1-d7/a1-a6,-(a7)

	lsl.w	#2,d0
        move.l  doorsos::heap,a6
	move.l	0(a6,d0.w),a6
;	_DEREF  d0,d6

        bra UnrelocFromPtrStart

UnrelocFromPtr:
        movem.l d1-d7/a1-a6,-(a7)
UnrelocFromPtrStart:
        tst.w   $8(a6)
        beq     UnrelocOK
        sub.w   #1,$8(a6)
        move.l  a6,a4
        tst.w   $8(a6)
        bne     UnrelocOK
        add.l   #$1a,a4
        move.w  (a4)+,d7
        move.w  d7,d6
        mulu.w  #10,d6
        move.l  a4,a3
        add.l   d6,a3
UnrelocLibLoop:
        tst.w   d7
        beq     UnrelocLibEnd
        move.l  a4,a0
        bsr     GetFileAddress
        tst.w   d0
        beq     UnrelocEnd
        move.w  d0,d1
        movem.l d1-d7/a0-a6,-(a7)
        move.l  a0,a6
        add.l   #2,a6
        bsr     UnrelocFromPtr
        movem.l (a7)+,d1-d7/a0-a6
        tst.w   d0
        bne     UnrelocLibEnd

	lsl.w	#2,d1
        move.l  doorsos::heap,a2
	move.l	0(a2,d1.w),a2


        add.l   #2,a2
        move.l  a2,a5
        clr.l   d0
        move.w  $16(a2),d0
        add.l   d0,a2
        move.w  #1,d3
        bsr     DoUnreloc
        add.l   #10,a4
        dbra    d7,UnrelocLibLoop
UnrelocLibEnd:
        tst.w   (a3)+
        beq     UnrelocNoROM
        move.l  $c8,a2
        sub.w   #2,a2
        move.w  #2,d3
        bsr     DoUnreloc
UnrelocNoROM:
        tst.w   (a3)+
        beq     UnrelocNoRAM
        clr.l   d0
        move.w  $18(a6),d0
        move.l  a6,a4
        add.l   d0,a4
;        move.l ($4),d0
;        and.l #$400000,d0
;        beq Unreloc89
        lea.l   _CALCULATOR(pc),a0
        move.b  #CALC_TI,(a0)
        lea.l   RAMTable92(pc),a2
        lea.l   2(a4),a4
;        bra UnrelocDoRAM
;Unreloc89:
;        lea.l _CALCULATOR(pc),a0
;        move.b #CALC_TI89,(a0)
;        lea.l RAMTable89(pc),a2
UnrelocDoRAM:
        move.w  #2,d3
        bsr     DoUnreloc
UnrelocNoRAM:
        clr.l   d0
        move.w  (a3)+,d0
        tst.w   d0
        beq     UnrelocRelEnd
        move.l  a6,a0
        add.l   d0,a0
        move.l  a6,d0
        sub.l   d0,(a0)
        bra     UnrelocNoRAM
UnrelocRelEnd:
        tst.w   $14(a6)
        beq     UnrelocOK
        add.l   #4,a3
        move.w  $12(a6),d0
        tst.w   d0
        beq     UnrelocOK
	lsl.w	#2,d0
        move.l  doorsos::heap,a1
	move.l	0(a1,d0.w),a1

UnrelocBSSLoop:
        clr.l   d0
        move.w  (a3)+,d0
        tst.w   d0
        beq     UnrelocBSSDone
        move.l  a6,a0
        add.l   d0,a0
        move.l  a1,d0
        sub.l   d0,(a0)
        bra     UnrelocBSSLoop
UnrelocBSSDone:
        move.w  $12(a6),d0
        tst.w   d0
        beq     UnrelocOK
        move.w  d0,-(a7)
        jsr     doorsos::HeapFree
        add.l   #2,a7
UnrelocOK:
        clr.l   d0
UnrelocEnd:
        movem.l (a7)+,d1-d7/a1-a6
        rts

DeallocBSS:
        tst.w   $12(a6)
        beq     NoBSS
        move.l  a4,-(a7)
        move.w  $12(a6),-(a7)
        jsr     doorsos::HeapFree
        add.l   #2,a7
        clr.w   $12(a6)
        move.l  (a7)+,a4
NoBSS:
        rts

;-----------------------------------------------------------------------------
; DoReloc		the universal relocation code
;			it is able to relocate every possible part
;			(ROM, RAM, ... addresses)
; Arguments		a2	relocationtable
;			a3	file relocation table
;			d3	multiplier (x*2^d3)
DoReloc:
        clr.l   d0
DoRelocStart:
        move.w  (a2)+,d5
        move.w  (a3)+,d6
DoRelocLoop:        
        clr.l   d4
        move.w  (a3)+,d4         ; index!
        move.l  a2,a1            ; from here, a1 is used for dereferencing
        clr.l   d2


        move.l  a1,a2
        cmp.w   #$8000,d4
        bcs     DoRelocNotWord
        moveq.l #1,d2
        and.w   #$7fff,d4
DoRelocNotWord:
        cmp.w   #$4000,d4
        bcs     DoRelocNotExtra
        move.l  a4,a1
        and.w   #$3fff,d4
        lsl.l   #2,d4
        add.l   d4,a1
        move.w  (a1),a1
        bra     DoRelocFuncLoop
DoRelocNotExtra:
        cmp.w   d5,d4
        bgt     DoRelocError
        lsl.l   d3,d4
        add.l   d4,a1
        cmp.w   #2,d3
        beq     DoRelocLoopDWordOfs
        move.w  (a1),a1
        add.l   a5,a1
        bra     DoRelocFuncLoop

DoRelocLoopDWordOfs:
        move.l  (a1),a1

DoRelocFuncLoop:
        clr.l   d4
        move.w  (a3)+,d4
        tst.w   d4
        beq     DoRelocFuncEnd
        move.l  a6,a0
        add.l   d4,a0
        move.l  a1,d4
        tst.l   d2
        bne     DoRelocAddWord
        tst.l   d0
        bne     DoRelocSubDWord

DoRelocAddDWord:
        add.l   d4,(a0)
        bra     DoRelocFuncLoop

DoRelocSubDWord:
        sub.l   d4,(a0)
        bra     DoRelocFuncLoop

DoRelocAddWord:
        tst.l   d0
        bne     DoRelocSubWord
        add.w   d4,(a0)
        bra     DoRelocFuncLoop

DoRelocSubWord:
        sub.w   d4,(a0)
        bra     DoRelocFuncLoop

DoRelocFuncEnd:
        dbra    d6,DoRelocLoop
        clr.l   d0
        rts				; Successful relocated

DoRelocError:				; Error happened
        moveq.l #1,d0
        rts

DoUnreloc:
        moveq.l #1,d0
        bra     DoRelocStart

; TODO:
; Searches for a file (used for libraries)
; Currently only in the main folder
; Input     a0		Pointer to the filename
GetFileAddress:
; main folder
        move.w  #$c,d0
;        GET_REL_RAM_WORD 0,-$18a,d0
	lsl.w	#2,d0
        move.l  doorsos::heap,a5
	move.l	0(a5,d0.w),a5
;        _DEREF d0,a5		; Put address of main folder in a5

        move.w	2(a5),d0	; Put number of files in d0
        add.l	#4,a5
        move.l	(a0),d1		; libraryname (first 4 letter)
        move.l	4(a0),d2	; (+ last 4 letter)
gfaFindLoop:
        tst.w	d0
        beq	gfaEnd
        cmp.l	(a5),d1
        bne	gfaNext
        cmp.l	4(a5),d2
        bne	gfaNext
        move.w	$a(a5),d0	; get file handle
        move.w	d0,d1

	lsl.w	#2,d1
        move.l  doorsos::heap,a0
	move.l	0(a0,d1.w),a0	; store fileaddress in a0
;        _DEREF d1,a0

        rts			; and return

gfaNext:           
        add.l	#$c,a5
        dbra	d0,gfaFindLoop
gfaEnd:
        clr.w	d0
        move.l	#0,a0		; file not found
        rts
