
;----------------------------------------------------------------------------
; Project		[tie]
; Module		kernel
;
; This kernel is originally written by Rusty Wagner which is used in
; PlusShell. Thanks to him for releasing of his kernel code.
;
; adapted by            MXM-TriX        mxmtrix@gmx.net
; released by           MXM Projects    http://mxm.ticalc.org
;
; Latest revision	2000/01/10
;----------------------------------------------------------------------------

;----------------------------------------------------------------------------
; The technique behind [tie]
; The [tie] boot program installs a ROM table at $CC with
; all supported ROM routines
; Structure:		dc.w	ROM_ROUTINE_NUMBER,
;			dc.l	ROM_ROUTINE_ADDRESS,
;			dc.w	ROM_ROUTINE_NUMBER,
;			dc.l	...
;			dc.w	$ffff		; END MARKER
; When a ROM routine isn't found, then an error is should be displayed
;----------------------------------------------------------------------------


; Since the  "doors"  did such a good work and [tie] is binary compatible,
; the program will use "doorsos.h", not a "tie.h"

			include	"doorsos.h"

                        xdef _ti92plus          ; Should be _ti92! (future)
			xdef _main

CALC_TI			equ	$92		; Flag to indicate the ti92

; The [tie]link program will detect these addresses and replace that stuff
; -> protected!
tios::FindSymEntry	equ	_RAM_CALL_F20
TIE_KERNEL_START	equ	_RAM_CALL_F01
TIE_KERNEL_SIZE		equ	_RAM_CALL_F02
ROMBASE			equ	_RAM_CALL_003

tios::ST_Flags2		equ	_RAM_CALL_010
tios::ST_Flags		equ	_RAM_CALL_016

tios::OSOnBreak         equ     doorsos::globals+$f02
tios::SYM_ENTRY.hVal	equ	10


doorsos::heap		equ	doorsos::globals+$1902


;--------------------------------------------------------------- BOOT PROGRAM

; The main function is the boot program (this code has to be at the start!)
_main:
BOOT_tie:
	clr.l	doorsos::globals+$1360  	; Clear eventhandler (else endlessloop)

	move.l	#doorsos::globals,a0
	lea	Logo(PC),a1
	move.w	#959,d0
CopyLogo:
	move.l	(a1)+,(a0)+
	dbra.w	d0,CopyLogo
	
	move.w	#$700,d0			; Prepare the following code
	trap	#1				; superuser needed!
	bclr.b	#2,($600001)			; Lower memory access

; Install ROM functions
	move.l	#$ce,a0
	move.l  a0,($C8)
	move.w	#$3cc,($cc)			; For the internal relocation routine
	lea	ROM_Functions(PC),a1
InstallROMFunctionssupport:
	clr.l	d1
	move.w	(a1)+,d1
	cmp.w	#$ffff,d1
	beq	ROMFunctionsinstalled
	lsl.w	#2,d1
	move.l	(a1)+,0(a0,d1.w)
	bra	InstallROMFunctionssupport
ROMFunctionsinstalled:


StartKernelInstall:

	move.w	#TIE_KERNEL_SIZE,d1		; Will be adapted by tielink
        move.l  #TIE_KERNEL_START,a1
        lea     __kernel_start(PC),a0
	move.l	a1,a2

	move.l  a1,doorsos::globals+$18f6	; Mark end of usable RAM

CopyKernel:
        move.l  (a0)+,(a1)+
	dbra.w	d1,CopyKernel

	lea	tie_onkey-__kernel_start(a2),a1
	move.l	a1,($78)

	
        lea     tie_timerfct-__kernel_start(a2),a1   	; Trap #0 redirection
	move.l	a1,($80)


	; Kernel information for user programs
	move.w	#$1000,($30)
	move.w	#'TY',($32)

	lea	user_exec-__kernel_start(a2),a1
	move.l	a1,($34)

	lea	Reloc-__kernel_start(a2),a1
	move.l	a1,($38)
	
	lea	RelocFromPtr-__kernel_start(a2),a1
	move.l	a1,($3C)

	lea	Unreloc-__kernel_start(a2),a1
	move.l	a1,($40)
	
	lea	UnrelocFromPtr-__kernel_start(a2),a1
	move.l	a1,($44)

	bset.b	#2,($600001)		; Protection of the lower memory
	trap	#1			; Old sr register

	; Now wait a little bit
Wait:
	tst.w	doorsos::kb_globals+$1c
	beq	Wait
	clr.w	doorsos::kb_globals+$1c
	; Restore stateline
	lea	doorsos::globals+121*30,a0
	move.w	#14,d0
RestoreStateLine:
	move.w	#-1,(a0)+
	dbra.w	d0,RestoreStateLine
	pea	update_stl(PC)
	jsr	doorsos::ST_showHelp
	lea	4(a7),a7
	rts

update_stl	dc.b "-",0			; update the statusline
	EVEN
Logo		incbin  "tielogo.bin"

	EVEN

;------------------------------------------------------------------ ROM TABLE
; The TI92 owns $3cc ROM functions
ROM_MARKER		dc.b    "ROM "		; Mark position for tielink
ROM_Functions		ds.b	$3cc*6

	EVEN
__kernel_start:
Marker			dc.b	"TY"

	EVEN

;---------------------------------------------------------------------- TRAPS
; Low level [tie] routines
tie_onkey:
	move.w	#$2700,sr
	move.l	a0,-(a7)
	lea	OnKey(PC),a0
	move.w	#1,(a0)
	move.l	(a7)+,a0
	move.l	ROMBASE+$78,-(a7)
	rts

tie_timerfct:
; Trap #0  -> necessary for catching calls from TIBasic!
	move.w	#$2700,sr

	lea	ProgCount(PC),a0
	tst.w	(a0)
	bne	original_trap#0

        ; tibasic call?
	cmp.w	#$154,d0
	beq	Check_Shift

        cmp.w   #$1d0,d0
        bne     original_trap#0

        ; catched a TI Basic call, now check if a assembly program has been
        ; called

        move.w  doorsos::globals+$1270,d0       ; get handle of the PRGM
	beq	No68k

        lea     Prgm(PC),a0
	move.w	d0,(a0)				; Store handle

        lsl.w   #2,d0                           ; get PRGM address
	move.l	doorsos::heap,a0
	move.l	0(a0,d0.w),a0
	move.l	6(a0),d0			; skip size (2 bytes) and bsr code (4 bytes)
	cmp.l	#'68kP',d0
	bne	No68k

	lea	Timer(PC),a0
	add.w	#1,(a0)
	cmp.w   #3,(a0)
	bne	No68k
	clr.w	(a0)

	; start the relocation routine in user mode
StartExec:
	move.l	usp,a0
	move.l	2(a7),-(a0)
	move.l	a0,usp
	lea	exec(PC),a0
	move.l	a0,2(a7)

	; Successfully done
	clr.w	d0
	rte

No68k:
        move.w  #$1d0,d0                        ; original TIOS call

original_trap#0:
	move.l	ROMBASE+$80,-(a7)		; original TIOS routine
	rts

Check_Shift:

	lea	ShiftFlag(PC),a0
	cmp.w	#0,(a0)
	bgt	WaitingforONKey

	btst.b	#2,tios::ST_Flags2+$f		; When the SHIFT is displayed
						; in the status line, then
						; SHIFT has been pressed
	beq	original_trap#0
	clr.w	OnKey-ShiftFlag(a0)		; Now the ON key is accepted
	move.w	#16,(a0)			; shift timer started
	bra	original_trap#0

WaitingforONKey:
	sub.w	#1,(a0)
	tst.w	OnKey-ShiftFlag(a0)
	beq	original_trap#0

	; Shift ON pressed!
	; find shell (in main)
	clr.w	OnKey-ShiftFlag(a0)		; clear <ON> key flag
	clr.w	(a0)				; clear shift flag

	movem.l	d1-d7/a1-a6,-(a7)
	bclr.b	#2,tios::ST_Flags2+1		; clear all shift flags 
	bclr.b	#2,tios::ST_Flags2+$f		; (logical)
	bclr.b	#2,tios::ST_Flags+3
	bset.b	#7,tios::ST_Flags+2
	jsr	doorsos::ST_eraseHelp		; actualize statusline

	lea	shellname(PC),a0	
	bsr	GetFileAddress
	movem.l (a7)+,d1-d7/a1-a6
	cmpa.l	#0,a0
	beq	ShellNotFound_msg

	; Start execution of shell
	lea	Prgm(PC),a0
	move.w	d0,(a0)

	bra	StartExec

ShellNotFound_msg:
	pea	shellnotfound(PC)
	jsr	doorsos::ST_showHelp
	lea	4(a7),a7
	move.w	#$154,d0
	move.l	ROMBASE+$80,-(a7)
	rts


;----------------------------------------------------------- CRASH PROTECTION
CrashHandler:
	; Set stack address, return address and signal that a crash happened
	move.w	#$2700,sr
	lea	ProgCount(PC),a0
	move.w	#1,CrashID-ProgCount(a0)
	lea	ProgramCrashed(PC),a1
	move.l	a1,2(a7)
	rte				; Return to crash handler


;-------------------------------------------------------------- USER ROUTINES
; int user_exec (int PrgmHandle)
; This is the exec for the user (address at $34)
user_exec:
	lea	Prgm(PC),a0
	move.w	d0,(a0)
	bra	exec
		

;-------------------------------------------------------- RELOCATION ROUTINES

        include "exec.h"

;------------------------------------------------------------------ VARIABLES
	EVEN
_CALCULATOR dc.b CALC_TI

	EVEN

RAMTable92 dc.w $16
           dc.l $92,240,128,_RAM_CALL_003,30,337,340,338,344,342,345,8192,3840,
           dc.l 16384,0,0,_RAM_CALL_010,_RAM_CALL_015+$1902,$b,$c,_RAM_CALL_F03,
           dc.l _RAM_CALL_015,_RAM_CALL_016

RAMTable89emu  dc.w $16
           dc.l 0,160,100,_RAM_CALL_003,30,337,340,338,344,342,345,8192,3840,
           dc.l 16384,0,0,_RAM_CALL_010,_RAM_CALL_015+$1902,$b,$c,_RAM_CALL_F03,
           dc.l _RAM_CALL_015,_RAM_CALL_016

ProgCount		dc.w	0
ShiftFlag		dc.w	0
Timer			dc.w	0
Prgm			dc.w	0
OnKey			dc.w	0
ScreenHandle		dc.w	0


; Anti Crash protection
CrashID			dc.w	0
UserExit		dc.l	0,0,0,0		; max. 4 programs called
CrashStack		dc.l	0,0,0,0

CrashTable		dc.b	$08,$0c,$10,$14,$18,$1c,$20,$2c,$24,$64+2*4,
			dc.b	$64+6*4,$80+3*4,$80+5*4,$80+6*4,$80+7*4,
			dc.b	$80+8*4,$80+11*4,$80+12*4,$80+13*4,$80+14*4,
			dc.b	$80+15*4,0

; Crash message
CrashHappened		dc.b	"[tie] program crashed",0

BadFileFormatMsg	dc.b	"[tie] program isn't valid",0

        EVEN

WrongLibVersionMsg      dc.b    "[tie] wrong lib:  "
WrongLibVersionName	dc.b	0,0,0,0,0,0,0,0,0

        EVEN
MissingLibMsg           dc.b    "[tie] lib not found:  "
MissingLibName		dc.b	0,0,0,0,0,0,0,0,0

	EVEN
NoMemMsg		dc.b	"[tie] out of memory",0

	EVEN
tietooold		dc.b	"[tie] kernel is too old!",0
	EVEN

shellnotfound		dc.b	"[tie] no shell found",0
        EVEN
libnotexecutable        dc.b    "[tie] library isn't executable",0

	EVEN
shellname		dc.b	"shell",0

__kernel_end:

	END
