
/*
 * Project      [tie]
 * Module       tielink
 * Author       TriX
 * Description  Kernel linker and backup linker
 * Latest revision     2000/01/10
 */

#include <stdio.h>
#include <stdlib.h>

#define uchar   unsigned char
#define ulong   unsigned long

#define TIE     getenv ("TIE")

int main (int argc, uchar *argv[]);

/* 
 * These prototypes are for writing some stuff to a destination!
 * The functions (Write****) will do an addition!
 */

void WriteDWord (uchar *destination, ulong summ);
void WriteWord (uchar *destination, ulong summ);
void WriteByte (uchar *destination, ulong summ);
ulong ReadDWord (uchar *source);
ulong ReadWord (uchar *source);
ulong ReadByte (uchar *source);


void WriteDWord (uchar *destination, ulong summ)
{
   uchar source[4];
   int overflow = 0, i;

   source[0] = (summ >> 24) & 0xff;
   source[1] = (summ >> 16) & 0xff;
   source[2] = (summ >> 8)  & 0xff;
   source[3] = (summ)       & 0xff;
   for (i = 3; i >= 0; i--)
   {
      if (destination[i] + source[i] + overflow > 0xff)
      {
         destination[i] = (destination[i] + source[i] + overflow) & 0xff;
         overflow = 1;
      } else {
         destination[i] = (destination[i] + source[i] + overflow) & 0xff;
         overflow = 0;
      } /* if */
   } /* for */
}

void WriteWord (uchar *destination, ulong summ)
{
   uchar source[2];
   int overflow = 0, i;

   source[0] = (summ >> 8) & 0xff;
   source[1] = (summ)      & 0xff;
   for (i = 1; i >= 0; i--)
   {
      if (destination[i] + source[i] + overflow > 0xff)
      {
         destination[i] = (destination[i] + source[i] + overflow) & 0xff;
         overflow = 1;
      } else {
         destination[i] = (destination[i] + source[i] + overflow) & 0xff;
         overflow = 0;
      } /* if */
   } /* for */
}

void WriteByte (uchar *destination, ulong source)
{
   destination[0] = (destination[0] + (source & 0xff)) & 0xff;
}

ulong ReadByte (uchar *source)
{
   return source[0];
}

ulong ReadWord (uchar *source)
{
   return (source[0] << 8) + (source[1]);
}

ulong ReadDWord (uchar *source)
{
   return (source[0] << 24) + (source[1] << 16) + (source[2] << 8) + (source[3]);
}


/* Function ReadFile:  Reads the whole file into memory */
int ReadFile (char* File, ulong *filesize, int offset, unsigned char **Pointer)
{
   FILE *tmp;
   int i;

   if ((tmp = fopen (File, "rb")) == NULL)
   {
     printf ("Can't open file %s\n", File);
     exit (-1);
   }
   fseek (tmp, 0, SEEK_END);
   *filesize = ftell (tmp);
   fseek (tmp, offset, SEEK_SET);
   *Pointer = (unsigned char*) malloc (*filesize);
   i = 0;
   while (!feof (tmp))
   {
      (*Pointer)[i] = fgetc (tmp);
      i++;
   } /* while */
   fclose (tmp);
   return 1;

} /* ReadFile */


/* Main function */
int main (int argc, uchar *argv[])
{
   char filename[128], kernelfile[128], tempname[128], tiosstr[32];
   uchar *kernel, *rom, *backup;
   ulong kernelsize, kernelstart, memorytop, kernelhigh, boot,
         tiekernelsize, romsize, backupsize, pos, offset, i, nr, DestDWord,
         k, UseWord, ROMVersion;
   short int chksum;
   ulong ROMTable[0x3ca] = {0};

   ulong kerneldestination;
   /* RAMfile addresses */

   /* 
    * tblInput marks the beginning of the EVENT caller, "ROM " is defined by
    * our kernel
    */
   uchar *tblInput = "tblInput", *ROMMarker = "ROM ";
   FILE *outfile;

   printf ("[tie] kernel linker 1.1c by MXM Projects\n");

   if (argc < 2 || argv[1][0] == '\0')
   {
      printf ("\nCalling options:    tielink <backupfile>\n\n");
      exit (0);
   } /* if */

   /* Find tios identification from the backup */
   ReadFile (argv[1], &backupsize, 0, &backup);

   /* Generate TIOS identification string */
   sprintf (tiosstr, "tios_%-7s", &backup[0x40]);

   /* HEAP top address */
   if (strcmp (&backup[0x40], "2.1") == 0)
      memorytop = 0x40000;
   else
      memorytop = 0x20000;

   /* Generate ROM version */
   ROMVersion = 0xf000 + ((backup[0x40] - '0') << 8) +
                         ((backup[0x42] - '0') << 4);

   if (backup[0x43] != '\0') ROMVersion += backup[0x43] - '0';

   sprintf (kernelfile, "%s/asm/kernel/tie.92p", TIE);
   ReadFile (kernelfile, &kernelsize, 0, &kernel);

   /* Find kernel high address */
   kernelhigh = 0x3ca * 4;  /* Just skip some bytes from the BOOT section */
   while (memcmp (kernel + kernelhigh, "TY", 2) != 0) kernelhigh++;

   kernelhigh -= 88;
   kernelstart = ReadWord (&kernel[88 + 2]);
   offset = ReadWord (&kernel[88 + 0xc]) + 88;
   boot = kernelhigh - kernelstart;
   tiekernelsize = kernelsize - kernelstart - boot + 4; /* It's saftey to add 4 bytes */
   kerneldestination = memorytop - ((tiekernelsize >> 2) << 2) - 4;

   printf ("Backupversion:   %s\n"
           "Boot section is: %ld\t bytes\n"
           "Kernel size:     %ld\t bytes\n"
           "Kernel start at: %x\n",
           &backup[0x40],
           boot,
           tiekernelsize,
           kerneldestination);


   /* Load TIOS table into the memory */
   sprintf (tempname, "%s/asm/kernel/romtbl.o", TIE);
   ReadFile (tempname, &romsize, 0, &rom);

   /* Start kernel relocation    */
   /* First start with ROM table */
   pos = 0;
   while (memcmp (rom + pos, tiosstr, 12) != 0 && pos < romsize) pos++;
   if (pos == romsize)
   {
      printf ("Your ROM version is not supported by [tie], please contact MXM Projects\n\n");
      exit (-1);
   } /* if */
   pos += 18;
   /* Copy ROM image */
   while (ReadWord (rom + pos) != 0xffff)
   {
      ROMTable[ReadWord (rom + pos)] =  ReadDWord (rom + pos + 2);
      if (memorytop == 0x20000 && ROMTable[ReadWord (rom + pos)] > 0x200000)
         ROMTable[ReadWord (rom + pos)] -= 0x200000;   /* ROM 1.x */
      pos += 6;
   } /* while */

   pos = 88 + 0x1a;             /* Start of relocation table */
   /* Overread the Import table */
   i = ReadWord (kernel + pos);
   if (i != 0)
   {
      printf ("Internal error 001\n\n");
      exit (1);
   }
   pos += 2;
   /* Start relocation (ROM functions) */
   if (ReadWord (kernel + pos) != 0)
   {
      nr = ReadWord (kernel + pos + 2);
      pos += 4;
      for (i = 0; i <= nr; i++)
      {
         DestDWord = ROMTable[ReadWord (kernel + pos)];
         if (DestDWord == 0)
         {
            printf ("Internal error 002\n\n");
            exit (2);
         }
         pos += 2;
         while (ReadWord (kernel + pos) != 0)
         {
            WriteDWord (&kernel[88 + ReadWord (kernel + pos)], DestDWord);
            pos += 2;
         } /* while */
         pos += 2;
      } /* for */
   } /* if */

   /* Do RAM relocation */
   if (ReadWord (kernel + pos) != 0)
   {
      nr = ReadWord (kernel + pos + 2);
      pos += 4;
      for (i = 0; i <= nr; i++)
      {
         k = ReadWord (kernel + pos);
         pos += 2;
         UseWord = (k & 0x8000);
         k &= 0x7fff;
         if (k == 0xf01)
            k = kerneldestination;
         else if (k == 0xf02)
            k = (tiekernelsize / 4) - 1;
         else if (k == 0xf03)
            k = ROMVersion;
         else
            k = ROMTable[k];
         while (ReadWord (kernel + pos) != 0)
         {
            if (UseWord)
               WriteWord (&kernel[88 + ReadWord (kernel + pos) & 0x7fff], k);
            else
               WriteDWord (&kernel[88 + ReadWord (kernel + pos)], k);
            pos += 2;
         } /* while */
         pos += 2;
      } /* for */
   } /* if */

   pos += 6;
   /* Do normal relocation */
   while (ReadWord (kernel + pos) != 0)
   {
      if (ReadWord (kernel + pos) & 0x8000)
         WriteWord (&kernel[88 + ReadWord (kernel + pos) & 0x7fff],
                    kerneldestination);
      else
         WriteDWord (&kernel[88 + ReadWord (kernel + pos)],
                    kerneldestination);
      pos += 2;
   } /* while */

   /* Insert new TIE kernel into the backup */
   /* Prepare boot section with ROM table   */
   printf ("Writing new backup... ");
   pos = 0;
   while (memcmp (kernel + pos, ROMMarker, 4) != 0) pos++;
   pos += 4;
   for (i = 0; i < 0x3ca; i++)
   {
      if (ROMTable[i] != 0)
      {
         WriteWord (&kernel[pos], i);
         WriteDWord (&kernel[pos + 2], ROMTable[i]);
         pos += 6;
      } /* if */
   } /* for */
   WriteWord (&kernel[pos], 0xffff);

   /* Search for the tblInput marker        */
   pos = 0;
   while (memcmp (backup + pos, tblInput, 8) != 0) pos++;

   pos += 16;
   /* Register event handler */
   /* Notice:  I hope this will work on a ROM version < 2.1 */
   /*  The following calculation would be  pos + LCD_ADDR + 0x1088 */
   if (memorytop == 0x40000)
   {
      /* ROM 2.1 */
      backup[0x02da] = ((pos + 0x57a8) >> 8) & 0xff;
      backup[0x02db] = ((pos + 0x57a8)     ) & 0xff;
   } else {
      /* ROM 1.x */
      backup[0x02da] = ((pos + 0x54c8) >> 8) & 0xff;
      backup[0x02db] = ((pos + 0x54c8)     ) & 0xff;
   } // if
   /* And now insert the kernel */
   for (i = offset; i < kernelsize; i++)
   {
       backup[pos + i - offset] = kernel[i];
   } /* for */

   /* Write backup with checksum */
   outfile = fopen (argv[1], "wb");
   if (outfile == NULL)
   {
      printf ("Can't open destination file during backupwriteback\n");
      exit (-10);
   } /* if */
   chksum = 0;
   for (i = 0; i < backupsize - 2; i++)
   {
      chksum += backup[i];
      fputc (backup[i], outfile);
   } /* fputc */
   /* Write checksum */
   fputc ((chksum >> 8) & 0xff, outfile);
   fputc (chksum & 0xff, outfile);

   fclose (outfile);
   printf ("Done\n");

   free (rom);
   free (kernel);
   free (backup);
   return 0;
} /* main */
